// ignore_for_file: public_member_api_docs,
// ignore_for_file: use_late_for_private_fields_and_variables

import 'dart:developer';

import 'package:fl_chart/fl_chart.dart';
import 'package:healthrian_ble_support_for_wearcardiot/healthrian_ble_support_for_wearcardiot.dart';
import 'package:healthrian_common_support/healthrian_common_support.dart';
import 'package:lzma/lzma.dart';

mixin Oxi {
  /// [maxBufferStreamOxi] decides a maximum length of buffered Stream.
  ///
  /// For 40 seconds buffer and 100 sampling rate,
  /// It uses 40 * 1000.
  int maxBufferStreamOxi = 40 * 100;

  /// [lengthBufferStreamOxi] decides a length of buffered Stream,
  /// after flushing buffer when reaching a maximum size.
  ///
  /// For 30 seconds buffer and 100 sampling rate,
  /// It uses 30 * 100.
  int lengthBufferStreamOxi = 30 * 100;

  /// [screenBufferStreamingOxi] decides a length of buffered Stream,
  /// which is shown to screen.
  ///
  /// For 5 seconds buffer and 100 sampling rate,
  /// It uses 5 * 100 by default.
  int screenBufferStreamingOxi = 5 * 100;

  /// [refreshBufferOxi] decides a length to be hidden after refreshed.
  ///
  /// For 0.5 seconds range to be hidden and 250 sampling rate,
  /// It use 100 ~/ 2 by default.
  int refreshBufferOxi = 100 ~/ 2;

  /// [isFilteredOxi] decides whether pulse oxi signal is filtered.
  bool isFilteredOxi = true;

  /// [isCompressedOxi] decides whether pulse oxi signal is compressed.
  ///
  /// If it is `true`, pulse oxi signal is decompressed by LZMA algorithm.
  bool isCompressedOxi = false;

  int _channelCountOxi = 2;
  int _sampleCountOxi = 30;

  int get _totalCountOxi => _channelCountOxi * _sampleCountOxi;

  bool _isPaused = false;

  bool _isMeasuring = false;

  /// [skipSampleOxi] is the number of samples to be skipped.
  int skipSampleOxi = 1;

  /// [skipPacketOxi] is the number of packet to be skipped.
  int skipPacketOxi = 1;
}

extension BLEOxi on BLEManager {
  /// [startMeasurementOxi] sends read_oxi_packet to device,
  ///  and then, it registers all oxi streams of [streams].
  Future<void> startMeasurementOxi() async {
    await _sendCommandReadOxi();
    await _getConfigurationOxi();

    // final isRunnable = await _checkValidationOxi();
    // if (!isRunnable) return;

    streams
      ..rawStreamOxi = _getStreamOxi().broadcast
      ..decompressedStreamOxi = _getDecompressedStreamOxi().broadcast
      ..recentStreamOxi = _getRecentStreamOxi().broadcast
      ..bufferedStreamOxi = _getBufferedStreamOxi().broadcast
      ..bufferedFilterStreamOxi = _getBufferedFilterStreamOxi().broadcast
      ..bufferedPlotStreamOxi = _getBufferedPlotStreamOxi().broadcast
      ..bufferedPlotRTStreamOxi = _getBufferedPlotRTStreamOxi().broadcast;

    _isPaused = false;
    _isMeasuring = true;
  }

  /// [pauseMeasurementOxi] pause measurements for Pulse Oxi.
  Future<void> pauseMeasurementOxi() async {
    _isPaused = !_isPaused;
  }

  /// [stopMeasurementOxi] sends stop_oxi_packet to device.
  Future<void> stopMeasurementOxi() async {
    _isMeasuring = false;
    final control = points[Characteristics.DEVICE_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetStopOxi}');
    await control.write(Commands.packetStopOxi);
  }

  Future<bool> _checkValidationOxi() async {
    final confirm = points[Characteristics.DEVICE_CONTROL_CONFIRM];
    if (confirm == null) return false;

    final confirmValue = await confirm.read();
    if (confirmValue.isEmpty) return false;
    if (confirmValue[0] == 2) return false;
    return true;
  }

  Future<void> _sendCommandReadOxi() async {
    final control = points[Characteristics.DEVICE_CONTROL_POINT];
    if (control == null) return;

    log('Writes a packet ${Commands.packetReadOxi}');
    await control.write(Commands.packetReadOxi);
  }

  Future<void> _getConfigurationOxi() async {
    final configuration = points[Characteristics.PULSE_OXIMETER_CONFIGURATION];
    if (configuration == null) return;

    final configurationValue = await configuration.read();

    if (configurationValue.length < 2) return;

    _channelCountOxi = configurationValue[0];
    _sampleCountOxi = configurationValue[1];
  }

  Stream<List<int>> _getStreamOxi() async* {
    final continuous = points[Characteristics.PULSE_OXIMETER_CONTINUOUS];
    if (continuous == null) return;
    await continuous.setNotifyValue(true);
    await for (final res in continuous.value) {
      if (!_isMeasuring || device == null) break;
      if (_isPaused) continue;
      yield res;
    }
  }

  Stream<List<int>> _getDecompressedStreamOxi() async* {
    final events = streams.rawStreamOxi;
    if (events == null) return;
    await for (final event in events) {
      yield isCompressedOxi ? lzma.decode(event) : event;
    }
  }

  Stream<List<double>> _getRecentStreamOxi() async* {
    final events = streams.rawStreamOxi;
    if (events == null) return;
    final res = <double>[];
    await for (final event in events) {
      if (event.isEmpty) continue;
      res.clear();
      for (var i = 0; i < _totalCountOxi; i++) {
        final raw = toDouble(
          event[i * 3 + 0],
          event[i * 3 + 1],
          event[i * 3 + 2],
        );
        res.add(raw);
      }
      yield res;
    }
  }

  Stream<Channels> _getBufferedStreamOxi() async* {
    final events = streams.recentStreamOxi;
    if (events == null) return;
    final res = List.generate(12, (index) => <double>[]);
    await for (final event in events) {
      for (var i = 0; i < event.length; i++) {
        if (i.isEven) {
          res[0].add(event[i]);
        } else {
          res[1].add(event[i]);
        }
      }
      if (res[1].length > maxBufferStreamOxi) {
        res[0].removeRange(0, res[0].length - maxBufferStreamOxi);
        res[1].removeRange(0, res[1].length - maxBufferStreamOxi);
      }
      yield Channels.fromList(res);
    }
  }

  Stream<Channels> _getBufferedFilterStreamOxi() async* {
    final events = streams.bufferedStreamOxi;
    if (events == null) return;

    await for (final event in events) {
      // ignore: lines_longer_than_80_chars
      for (var i = event.l2.length - _sampleCountOxi; i < event.l2.length; i++) {
        // event.l1[i] = _dataFilter.applyFilter(0, event.l1[i], true);
        // event.l2[i] = _dataFilter.applyFilter(1, event.l2[i], true);
      }
      yield event;
    }
  }

  Stream<PlotChannels> _getBufferedPlotStreamOxi() async* {
    // ignore: lines_longer_than_80_chars
    final events = isFilteredOxi ? streams.bufferedStreamOxi : streams.bufferedFilterStreamOxi;

    if (events == null) return;

    final resList = List.generate(12, (index) => <FlSpot>[]);
    final res = PlotChannels.fromList(resList, 0);
    var x = 0.0;

    await for (final event in events) {
      // ignore: lines_longer_than_80_chars
      for (var i = event.l2.length - _sampleCountOxi; i < event.l2.length; i++) {
        final j = i - event.l2.length + _sampleCountOxi;
        res.l1.add(FlSpot(x + j, event.l1[i]));
        res.l2.add(FlSpot(x + j, event.l2[i]));
      }

      if (res.l2.length > event.l2.length) {
        res.l1.removeRange(0, res.l1.length - event.l1.length);
        res.l2.removeRange(0, res.l2.length - event.l2.length);
      }

      yield res;
      x += _sampleCountOxi;
    }
  }

  Stream<PlotChannels> _getBufferedPlotRTStreamOxi() async* {
    final events = streams.bufferedPlotStreamOxi;

    if (events == null) return;

    double skip(double x) {
      final newX = x % screenBufferStreamingOxi;
      final newXSkip = newX - newX % skipSampleOxi;
      return newXSkip;
    }

    FlSpot f(FlSpot original) {
      return FlSpot(
        skip(original.x),
        original.y,
      );
    }

    final l1 = List<FlSpot>.generate(
      screenBufferStreamingOxi ~/ skipSampleOxi,
      (index) => FlSpot.nullSpot,
      growable: false,
    );

    final l2 = List<FlSpot>.generate(
      screenBufferStreamingOxi ~/ skipSampleOxi,
      (index) => FlSpot.nullSpot,
      growable: false,
    );

    final resList = List.generate(
      12,
      (index) {
        if (index == 0) return l1;
        if (index == 1) return l2;
        return <FlSpot>[];
      },
    );

    final res = PlotChannels.fromList(resList, 0);
    var x = 0;

    await for (final event in events) {
      // ignore: lines_longer_than_80_chars
      for (var i = event.l2.length - _sampleCountOxi; i < event.l2.length; i++) {
        final j = i - event.l2.length + _sampleCountOxi;

        if (skipSampleOxi != 1) {
          if ((x + j) % skipSampleOxi != 0) continue;
        }

        final index = skip(x + j.toDouble());

        res.l1[index ~/ skipSampleOxi] = f(event.l1[i]);
        res.l2[index ~/ skipSampleOxi] = f(event.l2[i]);

        if (refreshBufferOxi > 0) {
          final refreshedIndex = skip(x + j.toDouble() + refreshBufferOxi);

          res.l1[refreshedIndex ~/ skipSampleOxi] = FlSpot.nullSpot;
          res.l2[refreshedIndex ~/ skipSampleOxi] = FlSpot.nullSpot;
        }
      }

      x += _sampleCountOxi;

      if (skipSampleOxi == 1 || x ~/ _sampleCountOxi % skipSampleOxi == 0) {
        yield res.copyWith(
          pivot: (x % screenBufferStreamingOxi).toDouble(),
        );
      }
    }
  }

//
// Stream<List<List<double>>> _getBufferedStreamOxi() async* {
//   final events = streams.recentStreamOxi;
//   if (events == null) return;
//
//   final res = List<List<double>>.generate(
//     _channelCountOxi,
//     (index) => <double>[],
//   );
//
//   await for (final event in events) {
//     for (var index = 0; index < _totalCountOxi; index++) {
//       res[index % _channelCountOxi].add(event[index]);
//     }
//
//     for (final r in res) {
//       if (r.length > maxBufferStreamOxi) {
//         r.removeRange(0, r.length - lengthBufferStreamOxi);
//       }
//     }
//     yield res;
//   }
// }
//
// Stream<List<List<FlSpot>>> _getBufferedPlotStreamOxi() async* {
//   final events = streams.recentStreamOxi;
//   if (events == null) return;
//
//   var x = 0.0;
//   final res = List<List<FlSpot>>.generate(
//     _channelCountOxi,
//     (index) => <FlSpot>[],
//   );
//
//   await for (final event in events) {
//     for (var index = 0; index < _totalCountOxi; index++) {
//       res[index % _channelCountOxi].add(
//         FlSpot(
//           x + index ~/ _channelCountOxi,
//           event[index],
//         ),
//       );
//     }
//
//     for (final r in res) {
//       if (r.length > maxBufferStreamOxi) {
//         r.removeRange(0, r.length - lengthBufferStreamOxi);
//       }
//     }
//     yield res;
//
//     x += event.length ~/ _channelCountOxi;
//   }
// }
}
